from statsmodels import api as sm
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from dero.reg.reg import _create_reg_df_y_x_and_dummies, _estimate_handling_robust_and_cluster
from dero.reg.order import _set_regressor_order

def quantile_reg(df, yvar, xvars, q=0.5, robust=True, cluster=False, cons=True, fe=None, interaction_tuples=None,
        num_lags=0, lag_variables='xvars', lag_period_var='Date', lag_id_var='TICKER'):
    """
    Returns a fitted quantile regression. Takes df, produces a regression df with no missing among needed
    variables, and fits a regression model. If robust is specified, uses heteroskedasticity-
    robust standard errors. If cluster is specified, calculated clustered standard errors
    by the given variable.

    Note: only specify at most one of robust and cluster.

    Required inputs:
    df: pandas dataframe containing regression data
    yvar: str, column name of outcome y variable
    xvars: list of strs, column names of x variables for regression

    Optional inputs:
    q: float between 0 and 1. Quantile of dependent variable to estimate coefficients for
    robust: bool, set to True to use heterskedasticity-robust standard errors
    cluster: False or str, set to a column name to calculate standard errors within clusters
             given by unique values of given column name
    cons: bool, set to False to not include a constant in the regression
    fe: None or str or list of strs. If a str or list of strs is passed, uses these categorical
    variables to construct dummies for fixed effects.
    interaction_tuples: tuple or list of tuples of column names to interact and include as xvars
    num_lags: int, Number of periods to lag variables. Setting to other than 0 will activate lags
    lag_variables: 'all', 'xvars', or list of strs of names of columns to lag for regressions.
    lag_period_var: str, only used if lag_variables is not None. name of column which
                    contains period variable for lagging
    lag_id_var: str, only used if lag_variables is not None. name of column which
                    contains identifier variable for lagging

    Returns:
    If fe=None, returns statsmodels regression result
    if fe is not None, returns a tuple of (statsmodels regression result, dummy_cols_dict)
    """
    regdf, y, X, dummy_cols_dict, lag_variables = _create_reg_df_y_x_and_dummies(df, yvar, xvars, cluster=cluster, cons=cons, fe=fe,
                                                                  interaction_tuples=interaction_tuples, num_lags=num_lags,
                                                                  lag_variables=lag_variables, lag_period_var=lag_period_var,
                                                                  lag_id_var=lag_id_var)

    mod = sm.QuantReg(y, X)

    result = _estimate_handling_robust_and_cluster(regdf, mod, robust, cluster, q=q)

    # Only return dummy_cols_dict when fe is active
    if fe is not None:
        return result, dummy_cols_dict
    else:
        return result


def reg_for_each_quantile_output_plot(main_iv, *reg_args, num_quantiles=8, main_iv_label=None, outpath=None,
                                      clear_figure=True, **reg_kwargs):
    """
    Creates a plot of effect of main_iv on yvar at different quantiles. To be used after
    reg_for_each_quantile_produce_result_df

    :param main_iv: str, column name of independent variable of interest
    :param reg_args: see quantile_reg for args
    :param num_quantiles: number of quantile regressions to run. will be spaced evenly. higher numbers
                          produce smoother graphs but take longer to run
    :param main_iv_label: str, label of independent variable of interest
    :param outpath: str, filepath to output figure. must include matplotlib supported extension such as .pdf or .png
    :param clear_figure: bool, True wipe memory of matplotlib figure after running function
    :param reg_kwargs: see quantile_reg for kwargs
    :return: result_df from reg_for_each_quantile_produce_result_df
    """

    main_iv_label = _validate_quantile_reg_and_plot_args(main_iv, main_iv_label)

    result_df = reg_for_each_quantile_produce_result_df(
        main_iv,
        *reg_args,
        num_quantiles=num_quantiles,
        **reg_kwargs
    )

    quantile_plot_from_quantile_result_df(
        result_df,
        reg_args[1], #yvar
        main_iv_label,
        outpath=outpath,
        clear_figure=clear_figure
    )

    return result_df

def reg_for_each_quantile_produce_result_df(main_iv, *reg_args, num_quantiles=8, **reg_kwargs):
    """
    Produce result DataFrame of running multiple quantile regressions spaced out between the
    (0,1) interval

    :param main_iv: str, column name of independent variable of interest
    :param reg_args: see quantile_reg for args
    :param num_quantiles: number of quantile regressions to run. will be spaced evenly. higher numbers
                          produce smoother graphs but take longer to run
    :param reg_kwargs: see quantile_reg for kwargs
    :return:
    """
    quantiles = _create_quantiles_from_num_quantiles(num_quantiles)
    results = [(quantile_reg(*reg_args, q=q, **reg_kwargs), q) for q in quantiles]
    main_iv = _set_regressor_order([main_iv], reg_kwargs)[0]
    simple_results = [_produce_simplified_result_list(result[0], result[1], main_iv) for result in results]
    return pd.DataFrame(simple_results, columns=['q', 'a', 'b', 'lb', 'ub'])


def _produce_simplified_result_list(res, q, main_iv):

    # Handle possibility of dummy cols dict coming through with result as tuple
    if isinstance(res, tuple):
        res = res[0]

    return [q, res.params['const'], res.params[main_iv]] + \
           res.conf_int().ix[main_iv].tolist()


def _create_quantiles_from_num_quantiles(num_quantiles):
    quant_lower_bound = 1 / (num_quantiles + 1)
    quant_upper_bound = quant_lower_bound * num_quantiles + quant_lower_bound / 2
    return np.arange(quant_lower_bound, quant_upper_bound, quant_lower_bound)


def quantile_plot_from_quantile_result_df(result_df, yvar, main_iv, outpath=None, clear_figure=True):
    """
    Creates a plot of effect of main_iv on yvar at different quantiles. To be used after
    reg_for_each_quantile_produce_result_df

    :param result_df: pd.DataFrame, result from reg_for_each_quantile_produce_result_df
    :param yvar: str, label of dependent variable
    :param main_iv: str, label of independent variable of interest
    :param outpath: str, filepath to output figure. must include matplotlib supported extension such as .pdf or .png
    :param clear_figure: bool, True wipe memory of matplotlib figure after running function
    :return:
    """
    p1 = plt.plot(result_df.q, result_df.b, color='black', label='Quantile Regression Slope')
    p2 = plt.plot(result_df.q, result_df.ub, linestyle='dotted', color='black',
                  label='95% Confidence Interval Lower Bound')
    p3 = plt.plot(result_df.q, result_df.lb, linestyle='dotted', color='black',
                  label='95% Confidence Interval Upper Bound')
    plt.title(f'Effect of {main_iv} on {yvar}')
    plt.ylabel(r'Conditional Coefficient')
    plt.xlabel(f'Quantiles of {yvar}')
    plt.legend()

    if outpath:
        plt.savefig(outpath)
    else:
        plt.show()

    if clear_figure:
        plt.clf()


def _is_quantile_reg_str(reg_str):
    return reg_str in ('quantile', 'quant', 'q', 'quantile reg', 'quantile_reg', 'quant_reg', 'quantreg', 'qreg', 'quantile regression')

def _validate_quantile_reg_and_plot_args(main_iv, main_iv_label):
    main_iv_label = _set_main_iv_label(main_iv, main_iv_label)

    return main_iv_label

def _set_main_iv_label(main_iv, main_iv_label):
    """
    Set default main iv label as main iv
    """
    if main_iv_label is None:
        return main_iv
    else:
        return main_iv_label



import numpy as np

def hypothesis_test(result, col_coef_dict):
    """
    result: statsmodels RegressionResult or dero SimplifiedRegressionResult
    col_coef_dict: dict, keys are names of columns in DataFrame used for regression. Values are
                   coefficients to apply to those columns.
    """
    r_matrix = _create_r_matrix_for_hypothesis_test(result, col_coef_dict)
    return result.t_test(r_matrix, use_t=True)


def _create_r_matrix_for_hypothesis_test(result, col_coef_dict):
    """
    col_coef_dict: dict, keys are names of columns in DataFrame used for regression. Values are
                   coefficients to apply to those columns.
    """
    index_coef_dict = _replace_dict_keys_with_key_index_in_list(col_coef_dict, result.model.exog_names)
    r_matrix = _numpy_array_from_index_coef_dict(result, index_coef_dict)

    return r_matrix


def _numpy_array_from_index_coef_dict(result, index_coef_dict):
    arr = np.zeros(len(result.model.exog_names))
    _modify_array_from_index_coef_dict(arr, index_coef_dict)

    return arr


def _modify_array_from_index_coef_dict(arr, index_coef_dict):
    """
    Note: inplace
    """
    for index, coef in index_coef_dict.items():
        arr[index] = coef


def _replace_dict_keys_with_key_index_in_list(dict_, list_):
    return {list_.index(key): value for key, value in dict_.items()}


import posixpath

from dero.latex.models.mixins import StringAdditionMixin
from dero.latex.texgen import _include_graphics_str


class Graphic(StringAdditionMixin):

    def __init__(self, filepath, width=r'\linewidth'):
        self._set_path(filepath)
        self.width = width

    def __repr__(self):
        return f'<Graphic({self.filepath}, width={self.width})>'

    def __str__(self):
        return _include_graphics_str(self.source_path, self.width)

    def _set_path(self, filepath: str):
        from dero.latex.logic.pdf import _latex_valid_basename

        basename = _latex_valid_basename(filepath)
        source_path = posixpath.join('Sources', basename)

        self.filepath = filepath
        self.source_path = source_path






from .tools import _to_list_if_tuple

def create_interaction_variables(df, interaction_tuples):
    """
    Note: inplace
    """
    interaction_tuples = _to_list_if_tuple(interaction_tuples)

    interacted_variables = [_create_interaction_variable(df, tup) for tup in interaction_tuples]

    return interacted_variables

def delete_interaction_variables(df, interaction_tuples):
    """
    Note: inplace
    """
    for_delete = [_interaction_tuple_to_var_name(tup) for tup in interaction_tuples]
    df.drop(for_delete, axis=1, inplace=True)

def _create_interaction_variable(df, interaction_tuple):
    """
    Note: inplace
    """

    full_name = _interaction_tuple_to_var_name(interaction_tuple)

    # Create base for multiplication by each variable
    df[full_name] = 1

    for col in interaction_tuple:
        df[full_name] = df[full_name] * df[col]

    return full_name

def _interaction_tuple_to_var_name(interaction_tuple):
    return ' x '.join(interaction_tuple)

def _collect_variables_from_interaction_tuples(interaction_tuples):
    items = set()
    for tup in interaction_tuples:
        [items.add(item) for item in tup]
    return list(items)


from copy import deepcopy
from typing import Union, AnyStr, List

import pandas as pd

from dero.latex.logic.tools import _add_if_not_none
from dero.latex.models.mixins import ReprMixin
from dero.latex.table.models.data.valuestable import ValuesTable
from dero.latex.table.models.labels.label import Label
from dero.latex.table.models.labels.table import LabelTable, LabelCollection
from dero.latex.table.models.spacing.columntable import ColumnPadTable, CellSpacer
from dero.latex.table.models.table.section import TableSection
from dero.latex.table.logic.panels.topleft import _set_top_left_corner_labels

class DataTable(TableSection, ReprMixin):
    """
    Represents a subsection in a panel, but tracks row and column labels, which may be consolidated when
    assembled into Panels then a Table

    Use DataTable.from_df to create a DataTable from a pandas DataFrame.
    """
    repr_cols = ['values_table', 'column_labels', 'row_labels']

    def __init__(self, values_table: ValuesTable, column_labels: LabelTable=None, row_labels: LabelTable=None,
                 top_left_corner_labels: Union[Label, str] = None):
        self.values_table = values_table
        self.column_labels = column_labels
        self.row_labels = row_labels
        self.top_left_corner_labels = _set_top_left_corner_labels(top_left_corner_labels)

        self.should_add_top_left = (column_labels is not None) and (row_labels is not None)

    def __add__(self, other):
        if isinstance(other, DataTable):
            row_labels_match = _determine_match(self.row_labels, other.row_labels)
            if row_labels_match or other.row_labels is None:
                # if right table has same or None row labels, eliminate right row labels. Just add values
                values_table = _add_if_not_none(self.values_table, other.values_table)
                column_labels = _add_if_not_none(self.column_labels, other.column_labels)
            else:
                # if right table has unique row labels, absorb them into middle of values table
                values_table = _add_if_not_none(
                    self.values_table,
                    ValuesTable(other.row_labels.rows),
                    other.values_table
                )
                column_labels = _add_if_not_none(
                    self.column_labels,
                    other.top_left_corner_labels,
                    other.column_labels
                )

            row_labels = self.row_labels
        elif isinstance(other, ColumnPadTable):
            values_table = self.values_table + other if self.values_table is not None else None
            column_labels = self.column_labels + other if self.column_labels is not None else None
            row_labels = self.row_labels
        elif isinstance(other, TableSection):
            values_table = _add_if_not_none(self.values_table, other)
            column_labels = self.column_labels
            row_labels = self.row_labels
        else:
            raise ValueError(f'must add DataTable or TableSection to type {type(self)}. Got type {type(other)}')

        return DataTable(
            values_table=values_table,
            column_labels=column_labels,
            row_labels=row_labels,
            top_left_corner_labels=self.top_left_corner_labels
        )

    @property
    def rows(self):
        try:
            return self._rows
        except AttributeError:
            self._rows = self._create_rows()

        return self._rows

    # Following properties/setters exist to recreate rows if user overrides values table or labels

    @property
    def column_labels(self):
        return self._column_labels

    @column_labels.setter
    def column_labels(self, labels: LabelTable):
        self._column_labels = labels
        self._recreate_rows_if_created()

    @property
    def row_labels(self):
        return self._row_labels

    @row_labels.setter
    def row_labels(self, labels: LabelTable):
        self._row_labels = labels
        self._recreate_rows_if_created()

    @property
    def values_table(self):
        return self._values_table

    @values_table.setter
    def values_table(self, table: ValuesTable):
        self._values_table = table
        self._recreate_rows_if_created()

    def _create_rows(self):

        rows = []

        if self.column_labels is not None:
            if self.should_add_top_left:
                column_labels = self.top_left_corner_labels + self.column_labels
            else:
                column_labels = self.column_labels
            rows += column_labels.rows

        # need to add row labels inline with values table
        if self.row_labels is not None:
            assert len(self.row_labels.rows) == len(self.values_table.rows)
            for label_row, value_row in zip(self.row_labels.rows, self.values_table.rows):
                rows.append(label_row + value_row)
        # no row labels, add values rows as they are
        else:
            rows += self.values_table.rows

        return rows

    @classmethod
    def from_df(cls, df: pd.DataFrame, include_columns=True, include_index=False,
                extra_header: str=None, extra_header_underline=True,
                top_left_corner_labels: Union[LabelTable, LabelCollection, List[AnyStr], AnyStr] = None,
                **kwargs):
        """
        Use for the most fine-grained control in creating tables. Construct DataTables from
        pandas DataFrames, modify labels as needed, assemble them into Panels, then create a latex Table with
        Table.from_panel_list.

        :param df:
        :param include_columns:
        :param include_index:
        :param extra_header: extra multicolumn header to place over the existing column labels (or over values if
                             there are no column labels). Useful when placing multiple DataTables horizontally
                             in a Panel.
        :param extra_header_underline: whether to add an underline under the extra header, if the extre header
                                       was passed
        :param top_left_corner_labels: additional labels to place in the top left corner. pass a single string
                                       or a list of strings for convenience. a list of strings will be create labels
                                       which span the gap horizontally and go downwards, one label per row. pass
                                       LabelCollection or LabelTable for more control.
        :param kwargs: DataTable kwargs
        :return:
        """
        values_table = ValuesTable.from_df(df)

        if include_columns:
            column_label_table = LabelTable.from_df_index(df.columns)
        else:
            column_label_table = None

        if include_index:
            row_label_table = LabelTable.from_df_index(df.index).T
        else:
            row_label_table = None

        dt = cls(
            values_table,
            column_labels=column_label_table,
            row_labels=row_label_table,
            top_left_corner_labels=top_left_corner_labels,
            **kwargs,
        )

        if extra_header is not None:

            # set underline
            if extra_header_underline:
                underline = 0 # place an underline under the singular label
            else:
                underline = None # no underline

            # create multicolumn label
            label = Label(extra_header, span=values_table.num_columns)
            header = LabelCollection([label], underline=underline)
            if include_columns:
                # add to existing
                dt.column_labels.label_collections.insert(0, header)
            else:
                # create column labels as extra header
                dt.column_labels = LabelTable([header])

        return dt

def _determine_match(labels1: LabelTable, labels2: LabelTable):
    # handle equality for None
    if labels1 is None:
        if labels2 is None:
            return True
        else:
            return False
    elif labels2 is None:
        # labels 1 must not be None here
        return False

    # here, both are not None
    return labels1.matches(labels2)



import pandas as pd
from functools import partial
from statsmodels.api import OLS

class SimplifiedBase:
    def _set_attrs(self, attr_dict):
        for attr in attr_dict:
            setattr(self, attr, attr_dict[attr])

class SimplifiedRegressionResult(SimplifiedBase):
    direct_attrs = [
            'params', 'pvalues', 'tvalues', 'nobs', 'rsquared_adj', 'bse', 'conf_int',
            'normalized_cov_params', 'cov_params_default', 'scale', 'cov_params',
            't_test'
    ]
    model_attrs = ['exog_names', 'endog_names']


    def __init__(self, **kwargs):
        _validate_attrs(kwargs, self.direct_attrs + self.model_attrs)
        model_kwargs = SimplifiedRegressionResult.pop_model_attrs(kwargs)
        self._set_attrs(kwargs)
        self.model = SimplifiedModel(**model_kwargs)

    @classmethod
    def from_statsmodels_result(cls, result):

        # Get direct attributes
        result_dict = _extract_attrs_into_dict(result, cls.direct_attrs)

        # Get attributes of model
        result_dict.update(_extract_attrs_into_dict(result.model, cls.model_attrs))

        return cls(**result_dict)

    @classmethod
    def pop_model_attrs(cls, attr_dict):
        """
        Note: pops from attr_dict inplace
        """
        outdict = {}
        for attr in attr_dict:
            if attr in cls.model_attrs:
                outdict[attr] = attr_dict[attr]

        # Must pop separately as cannot change size of iterating dict
        [attr_dict.pop(attr) for attr in outdict]
        return outdict


class SimplifiedModel(SimplifiedBase):

    def __init__(self, **kwargs):
        self._set_attrs(kwargs)

class UnsupportedResultAttributeException(Exception):
    pass

def _validate_attrs(attr_dict, valid_attrs):
    for attr in attr_dict:
        if attr not in valid_attrs:
            raise UnsupportedResultAttributeException(f'Attribute {attr} not supported for SimplifiedRegressionResult')

def _extract_attrs_into_dict(obj, attrs):
    result_dict = {}

    # Get direct attributes
    for attr in attrs:
        value = getattr(obj, attr)
        if isinstance(value, (pd.Series, pd.DataFrame, list, dict)):
            value = value.copy()
        result_dict[attr] = value

    return result_dict


def remove_lag_names_from_reg_results(reg_list, lags=(1,)):
    """
    Note: partially inplace
    """
    out_reg_list = []
    for ambiguous_result in reg_list:

        # Determine type of result
        if isinstance(ambiguous_result, tuple):  # Tuple of result, fe_dict
            result = ambiguous_result[0]
        else:  # just a single result
            result = ambiguous_result

        # Actually replace names
        result = _remove_lag_name_from_reg_result(result, lags=lags)

        # Add to output, depending on type of result
        if isinstance(ambiguous_result, tuple):  # Tuple of result, fe_dict
            out_reg_list.append((result, ambiguous_result[1]))
        else:  # just a single result
            out_reg_list.append(result)

    return out_reg_list


def _remove_lag_name_from_reg_result(result, lags=(1,)):
    """
    Note: partially inplace
    """
    result = SimplifiedRegressionResult.from_statsmodels_result(result)

    # Modify base properties inplace
    [
        _remove_lag_names_from_ambiguous_property(getattr(result, item), lags=lags) for item in (
            'params', 'pvalues', 'tvalues', 'bse', 'normalized_cov_params'
        )
    ]

    # Modify model properties and reassign (functions not inplace)
    for attr in ['endog_names', 'exog_names']:
        setattr(
            result.model,
            attr,
            _remove_lag_names_from_ambiguous_property(
                getattr(result.model, attr),
                lags=lags)
        )

    return result

def _remove_lag_names_from_ambiguous_property(ambiguous, lags=(1,)):
    """
    Note: Series and DataFrame operations inplace, str and list operations not inplace
    """
    if isinstance(ambiguous, pd.DataFrame):
        lag_func = partial(_remove_lag_names_from_df_index_and_columns, ambiguous)
    elif isinstance(ambiguous, pd.Series):
        lag_func = partial(_remove_lag_names_from_series_index, ambiguous)
    elif isinstance(ambiguous, str):
        lag_func = partial(_remove_lag_names_from_varname, ambiguous)
    elif isinstance(ambiguous, list):
        lag_func = partial(_remove_lag_names_from_list, ambiguous)
    else:
        raise ValueError(f'Must pass DataFrame, Series, str, or list. Got type {type(ambiguous)}')

    return lag_func(lags=lags)

def _remove_lag_names_from_df_index_and_columns(df, lags=(1,)):
    """
    Note: inplace
    """
    [_remove_one_lag_names_from_df_index_and_columns(df, num_lags=num_lags) for num_lags in lags]


def _remove_lag_names_from_series_index(series, lags=(1,)):
    """
    Note: inplace
    """
    [_remove_one_lag_names_from_series_index(series, num_lags=num_lags) for num_lags in lags]


def _remove_one_lag_names_from_df_index_and_columns(df, num_lags=1):
    """
    Note: inplace
    """
    rename_dict = {col: lag_varname_to_varname(col, num_lags=num_lags) for col in df.index}
    df.index = df.index.to_series().replace(rename_dict)
    df.columns = df.columns.to_series().replace(rename_dict)

def _remove_one_lag_names_from_series_index(series, num_lags=1):
    """
    Note: inplace
    """
    rename_dict = {col: lag_varname_to_varname(col, num_lags=num_lags) for col in series.index}
    series.index = series.index.to_series().replace(rename_dict)

def _remove_lag_names_from_list(list_, lags=(1,)):
    for lag in lags:
        list_ = _remove_one_lag_names_from_list(list_, lag)
    return list_

def _remove_one_lag_names_from_list(list_, num_lags=1):
    return [lag_varname_to_varname(item, num_lags=num_lags) for item in list_]

def _remove_lag_names_from_varname(varname, lags=(1,)):
    for lag in lags:
        varname = lag_varname_to_varname(varname, lag)
    return varname

def lag_varname_to_varname(varname, num_lags=1):
    return varname.replace(rf'_{{t - {num_lags}}}', '')



import pandas as pd

def suppress_controls_in_summary_df(summ_df, regressor_order, dummy_col_dicts, info_dict):
    regressors, fe, info = _each_row_set_to_keep_from_regressor_order_dummy_col_dicts_and_info_dict(
        regressor_order, dummy_col_dicts, info_dict
    )
    variables_df, fe_df, info_df = _split_summ_df_into_variables_fixed_effects_info(summ_df, fe, info)
    regressors_df, controls_df = _split_variable_df_into_regressors_and_controls(variables_df, regressors)
    controls_row = _create_controls_row_as_df(controls_df)
    return _combine_dfs(regressors_df, controls_row, fe_df, info_df)


def _all_rows_to_keep_from_regressor_order_dummy_col_dicts_and_info_dict(regressor_order, dummy_col_dicts, info_dict):
    regressors, fe, info = _each_row_set_to_keep_from_regressor_order_dummy_col_dicts_and_info_dict(
        regressor_order, dummy_col_dicts, info_dict
    )
    return regressors + fe + info


def _each_row_set_to_keep_from_regressor_order_dummy_col_dicts_and_info_dict(regressor_order, dummy_col_dicts,
                                                                             info_dict):
    regressors = regressor_order.copy()
    fe = _extract_all_fe_names_from_dummy_col_dicts(dummy_col_dicts)
    info = [col for col in info_dict]
    return regressors, fe, info


def _extract_all_fe_names_from_dummy_col_dicts(dummy_col_dicts):
    cols = []
    # Loop through models, the dummy variable dictionary for each
    for dummy_dict in dummy_col_dicts:
        # Will come through as None instead of dict if no fixed effects for this model
        if not dummy_dict:
            continue
        for col in dummy_dict:
            full_name = col + ' Fixed Effects'
            if full_name not in cols:
                cols.append(full_name)

    return cols


def _split_summ_df_into_variables_fixed_effects_info(df, fe, info):
    fe_mask = df.index.isin(fe)
    fe_df = df.loc[fe_mask]
    info_mask = df.index.isin(info)
    info_df = df.loc[info_mask]

    variables_mask = ~fe_mask & ~info_mask
    variables_df = df.loc[variables_mask]

    return variables_df, fe_df, info_df


def _split_variable_df_into_regressors_and_controls(variables_df, regressors, stderr=True):
    controls = [row for row in variables_df.index if row not in regressors + ['']]
    all_variables = regressors + controls

    if stderr:
        # Assign index values of name of regressor to stderr rows, that way both coef and stderr
        # are referenced by the same index
        f = lambda idx: sum([[x, x] for x in idx], [])
        variables_df.index = f(all_variables)

    # Select df of just regressors, with modified index
    regressors_mask = variables_df.index.isin(regressors)
    regressors_df = variables_df[regressors_mask]
    controls_df = variables_df[~regressors_mask]

    if stderr:
        # Reset indices so that stderrs have blank indices again
        f = lambda idx: sum([[x, ''] for x in idx], [])
        regressors_df.index = f(regressors)
        controls_df.index = f(controls)

    return regressors_df, controls_df


def _create_controls_row_as_df(controls_df):
    bool_df = pd.DataFrame(controls_df.apply(lambda x: bool(x.any()), axis=0)).T
    yes_no_df = bool_df.applymap(lambda x: 'Yes' if x else 'No')
    yes_no_df.index = ['Controls']

    return yes_no_df


def _combine_dfs(regressors_df, controls_row, fe_df, info_df):
    df_list = [regressors_df, controls_row, fe_df, info_df]
    return pd.concat(df_list, axis=0)




from dero.latex.models import Environment
from dero.latex.models.mixins import StringAdditionMixin
from dero.latex.texgen import _basic_item_str, _multi_option_item_str

class Item(StringAdditionMixin):

    def __init__(self, name, contents, pre_env_contents=None, post_env_contents=None, env_modifiers=None):
        self.env = Environment(name, modifiers=env_modifiers)
        self.contents = contents

        self._output = ''
        if pre_env_contents:
            self._output += pre_env_contents
        self._output += self.env.wrap(str(self.contents))
        if post_env_contents:
            self._output += post_env_contents

    def __repr__(self):
        return f'<Item(name={self.env.name}, contents={self.contents})>'

    def __str__(self):
        return self._output

class SimpleItem(StringAdditionMixin):

    def __init__(self, name, contents):
        self.name = name
        self.contents = contents

    def __repr__(self):
        return f'<{self.name.title()}({self.contents})>'

    def __str__(self):
        return _basic_item_str(self.name, self.contents)

class MultiOptionSimpleItem(StringAdditionMixin):

    def __init__(self, name, *options):
        self.name = name
        self.options = options

    def __repr__(self):
        return f'<{self.name.title()}({self.options})>'

    def __str__(self):
        return _multi_option_item_str(self.name, *self.options)


import statsmodels.api as sm


def get_model_class_by_string(model_string):
    if _is_logit_str(model_string):
        return sm.Logit
    elif _is_probit_str(model_string):
        return sm.Probit
    elif _is_ols_str(model_string):
        return sm.OLS
    else:
        raise ValueError(f'model string does not signify ols, probit, or logit. got {model_string}')


def _is_probit_str(model_string):
    return model_string.lower().strip() in ('probit','p','prob','prbt')

def _is_logit_str(model_string):
    return model_string.lower().strip() in ('logit','l','log','lgt')

def _is_ols_str(model_string):
    return model_string.lower().strip() in ('ols','o','reg','least squares', 'ordinary least squares')


from typing import Union, List

from dero.latex.figure.models.subfigure import Subfigure, Graphic
from dero.latex.models import Item
from dero.latex.models.caption import Caption
from dero.latex.models.label import Label
from dero.latex.logic.builder import build_figure_content
from dero.latex.models.document import Document

SubfigureOrGraphic = Union[Subfigure, Graphic]
SubfiguresOrGraphics = List[SubfigureOrGraphic]

class Figure(Item):
    """
    used for creating latex figures from images. Currently the main usage is the Figure class created with the method
    Figure.from_dict_of_names_and_filepaths. Pass a dictionary where the keys are names for subfigures and the values
    are filepaths where the image for the subfigure is located.

    """
    name = 'figure'

    def __init__(self, subfigures: SubfiguresOrGraphics, caption=None, label=None, centering=True, position_str=None):
        self.subfigures = subfigures
        self.caption = Caption(caption) if caption else None
        self.label = Label(label) if label else None
        self.centering = centering

        self._remove_subfigure_elevate_contents_to_figure_if_single_subfigure()

        content = build_figure_content(
            self.subfigures,
            caption=self.caption,
            label=self.label,
            centering=self.centering,
            position_str=position_str
        )

        super().__init__(self.name, content)

    def __repr__(self):
        return f'<Figure(subfigures={self.subfigures}, caption={self.caption})>'

    def __iter__(self):
        for subfigure in self.subfigures:
            yield subfigure

    def __getitem__(self, item):
        return self.subfigures[item]

    def as_document(self, landscape=False):
        from dero.latex.models.package import Package

        simple_package_strs = [
            'caption',
            'subcaption',
            'graphicx',
            'pdflscape'
        ]
        simple_packages = [Package(str_) for str_ in simple_package_strs]

        packages = simple_packages + [
            Package('geometry', modifier_str='margin=0.1in')
        ]

        return Document(packages, self, landscape=landscape)

    def to_pdf_and_move(self, as_document=True, outfolder: str=None, outname: str=None,
                        landscape=False):
        from dero.latex.logic.pdf import _document_to_pdf_and_move

        to_output: Figure = self
        if as_document:
            to_output: Document = self.as_document(landscape=landscape)

        if outfolder is None:
            outfolder = '.'
        if outname is None:
            outname = 'figure'

        _document_to_pdf_and_move(
            to_output,
            outfolder,
            image_paths=self.filepaths,
            outname=outname,
            as_document=as_document
        )

    @property
    def filepaths(self):
        filepaths = []
        for subfigure in self:
            if isinstance(subfigure, Subfigure):
                filepaths.append(subfigure.graphic.filepath)
            elif isinstance(subfigure, Graphic):
                filepaths.append(subfigure.filepath)
            else:
                raise ValueError(f'must pass Subfigures or Graphics to Figure. got type {type(subfigure)}')
        return filepaths

    @property
    def source_paths(self):
        source_paths = []
        for subfigure in self:
            if isinstance(subfigure, Subfigure):
                source_paths.append(subfigure.graphic.source_path)
            elif isinstance(subfigure, Graphic):
                source_paths.append(subfigure.source_path)
            else:
                raise ValueError(f'must pass Subfigures or Graphics to Figure. got type {type(subfigure)}')
        return source_paths

    @classmethod
    def from_dict_of_names_and_filepaths(cls, filepath_name_dict: dict, figure_name: str=None):
        """

        :param filepath_name_dict: dictionary where keys are names of subfigures and values
                                   are the filepaths to the images for those figures.
        :param figure_name: name for overall figure
        :return: Figure
        """
        subfigures = []
        for name, filepath in filepath_name_dict.items():
            subfigures.append(
                Subfigure(filepath, caption=name)
            )

        return cls(subfigures, caption=figure_name)

    def _remove_subfigure_elevate_contents_to_figure_if_single_subfigure(self):
        """
        If there is a single subfigure, leaving in subfigure format results in very small output. Must strip
        out subfigure layer, leaving only a figure with a graphic, then it can fill the page.
        :return:
        """
        if len(self.subfigures) != 1:
            return

        self.subfigures: List[Subfigure]
        # update width to whole page
        graphic = Graphic(self.subfigures[0].graphic.filepath, width=r'1.26\paperwidth')
        # need to turn off centering to cover whole page
        self.centering = False

        self.subfigures = [graphic]






import pandas as pd

def add_fixed_effects_rows(summ_df, fixed_effect_dict):
    """
    summ_df: regression summary df where models are columns
    fixed_effect_dict: dictionary where keys are names of fixed effects and values are a single
    boolean for whether to include the fixed effect for all models, or a list of booleans for
    whether to include the fixed effect for each model, e.g. for a 3 model summ_df:
        fixed_effect_dict = {
            'Industry': True,
            'Time': [False, False, True]
        }
    """

    fe_dict = _standardize_fixed_effects_dict(summ_df, fixed_effect_dict)

    for fixed_effect_name, booleans in fe_dict.items():
        summ_df = _add_fixed_effects_row(summ_df, booleans, fixed_effect_name)

    return summ_df


def _standardize_fixed_effects_dict(summ_df, fixed_effect_dict):
    num_models = len(summ_df.columns)

    out_dict = {}
    for fixed_effect_name, booleans in fixed_effect_dict.items():
        # Here we are being passed a list of booleans matching the size of models.
        # This is the correct format so just output
        if (isinstance(booleans, list)) and (len(booleans) == num_models):
            out_booleans = booleans

            # Here we are being passed a single boolean or a list with single boolean
        # Need to expand to cover all models
        else:
            if (not isinstance(booleans, list)):
                booleans = [booleans]
            if len(booleans) > 1:
                raise ValueError(
                    f'Incorrect shape of booleans for {fixed_effect_name} fixed effect passed. Got {len(booleans)} bools, was expecting {num_models}')
            out_booleans = [booleans[0]] * num_models

            # Final input checks
        assert isinstance(out_booleans, list)
        assert all([isinstance(b, bool) for b in out_booleans])
        assert isinstance(fixed_effect_name, str)

        # Assign to output dictionary
        out_dict[fixed_effect_name] = out_booleans

    return out_dict


def _add_fixed_effects_row(summ_df, bool_list, fixed_effect_name='Industry'):
    fixed_effects_row = pd.DataFrame([_get_yes_no(bool_) for bool_ in bool_list], columns=[fixed_effect_name + ' Fixed Effects']).T
    fixed_effects_row.columns = summ_df.columns
    return pd.concat([summ_df, fixed_effects_row], axis=0)

def _get_yes_no(bool_):
    if bool_:
        return 'Yes'
    else:
        return 'No'


from .iter import (
    reg_for_each_combo,
    reg_for_each_xvar_set,
    reg_for_each_combo_select_and_produce_summary,
    reg_for_each_xvar_set_and_produce_summary,
    reg_for_each_yvar,
    reg_for_each_yvar_and_produce_summary,
    reg_for_each_lag,
    reg_for_each_lag_and_produce_summary
)
from .reg import reg
from .summarize import produce_summary
from .select import select_models
from .args import RegressionSetArgs
from dero.reg.quantile import quantile_reg




import pandas as pd
import matplotlib.pyplot as plt

from dero.reg.interact import _interaction_tuple_to_var_name
from dero.reg.iter import _set_interaction_tuples
from dero.reg.hypothesis.lincom import hypothesis_test

# TODO: make module flexible to not having interaction

def interacted_lag_plot_from_reg_result_list(results, lag_tuple, main_iv, interaction_tuples, yvar,
                                             interaction_var_value=13.27, date_var='Year',
                                             outpath=None, clear_figure=True):
    plot_df = produce_simplified_result_df(
        results, lag_tuple, main_iv, interaction_tuples, interaction_var_value=interaction_var_value
    )
    return interacted_lag_plot_from_lag_result_df(
        plot_df, yvar, main_iv, date_var=date_var, outpath=outpath, clear_figure=clear_figure
    )

def interacted_lag_plot_from_lag_result_df(result_df, yvar, main_iv, date_var='Year', outpath=None, clear_figure=True):
    """
    Creates a plot of effect of main_iv on yvar at different quantiles. To be used after
    reg_for_each_quantile_produce_result_df

    :param result_df: pd.DataFrame, result from reg_for_each_quantile_produce_result_df
    :param yvar: str, label of dependent variable
    :param main_iv: str, label of independent variable of interest
    :param outpath: str, filepath to output figure. must include matplotlib supported extension such as .pdf or .png
    :param clear_figure: bool, True wipe memory of matplotlib figure after running function
    :return:
    """
    p1 = plt.plot(result_df.q, result_df.b, color='black', label='Lag Regression Slope')
    p2 = plt.fill_between(result_df.q, result_df.ub, result_df.lb, color='gray',
                          label='95% Confidence Interval Bound')
    plt.title(f'Effect of {main_iv} on {yvar} over Time')
    plt.ylabel(r'Lag Coefficient')
    plt.xlabel(f'Number of {date_var}s Lagged')
    plt.legend()

    if outpath:
        plt.savefig(outpath)
    else:
        plt.show()

    if clear_figure:
        plt.clf()

def produce_simplified_result_df(results, lag_tuple, main_iv, interaction_tuples, interaction_var_value=13.27):
    simple_results = _produce_simplified_results(
        results, lag_tuple, main_iv, interaction_tuples, interaction_var_value=interaction_var_value
    )
    return pd.DataFrame(simple_results, columns=['q', 'a', 'b', 'lb', 'ub'])


def _produce_simplified_results(results, lag_tuple, main_iv, interaction_tuples, interaction_var_value=13.27):
    interaction_tuples = _set_interaction_tuples(interaction_tuples, len(results))
    return [
        _produce_simplified_result_list_from_one_result(
            result, lag_tuple[i], main_iv, interaction_tuples[i], interaction_var_value=interaction_var_value
        ) for i, result in enumerate(results)
    ]


def _produce_simplified_result_list_from_one_result(res, lag, main_iv, interaction_tuple, interaction_var_value=13.27):
    # Handle possibility of dummy cols dict coming through with result as tuple
    if isinstance(res, tuple):
        res = res[0]

    interaction_name = _interaction_tuple_to_var_name(interaction_tuple)

    # Just reassign names for brevity
    b_iv = res.params[main_iv]
    b_interact = res.params[interaction_name]

    # 95% confidence interval +/- amount
    # Get standard error of linear combination
    col_coef_dict = {
        main_iv: 1,
        interaction_name: interaction_var_value
    }
    hypothesis_result = hypothesis_test(res, col_coef_dict=col_coef_dict)

    return [
        lag,
        res.params['const'],
        hypothesis_result.effect,
        hypothesis_result.conf_int()[0][0],  # lower 95% confidence interval
        hypothesis_result.conf_int()[0][1],  # upper 95% confidence interval
    ]


from dero.latex.logic.tools import show_contents

class StringAdditionMixin:

    def __add__(self, other):
        return str(self) + str(other)

    def __radd__(self, other):
        return str(other) + str(self)

    def join(self, iterable):
        return self.__str__().join(str(i) for i in iterable)


class ReprMixin:
    repr_cols = []

    def __repr__(self):
        if self.repr_cols:
            repr_col_strs = [f'{col}={getattr(self, col, None).__repr__()}' for col in self.repr_cols]
            repr_col_str = f'({", ".join(repr_col_strs)})'
        else:
            repr_col_str = ''
        return f'<{type(self).__name__}{repr_col_str}>'

    @property
    def readable_repr(self):
        return show_contents(self)


from typing import Union, Iterable

from dero.latex.table.models.data.dataitem import DataItem
from dero.latex.table.models.labels.collection import LabelCollection
from dero.latex.table.models.labels.label import Label
from dero.latex.table.models.spacing.cell import CellSpacer
from dero.latex.table.models.table.rowbase import RowBase
from dero.latex.texparser.clean import _remove_backslashes


class Row(RowBase):
    repr_cols = ['values']

    def __init__(self, values: Union[Iterable[DataItem], LabelCollection, Label]):

        # Don't allow nested rows. If the only values passed to a Row are a Row, then use the values of that
        # row rather than the row itself as values
        if isinstance(values, Row):
            self.values = values.values
        else:
            self.values = values

    def __str__(self):
        str_list = []
        for value in self.values:
            # Handle other forms of blanks
            if isinstance(value, Iterable) and len(value) == 1 and str(value[0]).strip() == '':
                str_list.append(' ')
            # The usual case, just join string representation of items
            else:
                str_list.append(str(value))

        return ' & '.join(str_list)

    @classmethod
    def from_latex_row_str(cls, latex_row_str):
        latex_row_str = _remove_backslashes(latex_row_str)
        str_values = latex_row_str.split(' & ')
        values = [DataItem(value) for value in str_values]
        return cls(values)


from .reg import reg
from ..ext_pandas.filldata import add_missing_group_rows, drop_missing_group_rows
from dero.reg.lag.create import _is_special_lag_keyword


def diff_reg(df, yvar, xvars, id_col, date_col, difference_lag=1, diff_cols=None, **reg_kwargs):

    if diff_cols == None:
        # All by default
        diff_cols = [yvar] + xvars

    df = create_differenced_variables(df, diff_cols, id_col=id_col, date_col=date_col, difference_lag=difference_lag)

    # Convert names in lists of variables being passed to reg
    reg_yvar, reg_xvars = _convert_variable_names(yvar, xvars, diff_cols)
    this_reg_kwargs = reg_kwargs.copy()
    if 'interaction_tuples' in reg_kwargs:
        this_reg_kwargs['interaction_tuples'] = _convert_interaction_tuples(reg_kwargs['interaction_tuples'], diff_cols)
    if 'lag_variables' in reg_kwargs:
        this_reg_kwargs['lag_variables'] = _convert_list_of_variables_to_difference_names(reg_kwargs['lag_variables'], diff_cols)


    result = reg(df, reg_yvar, reg_xvars, **this_reg_kwargs)

    differenced_names = [col + ' Change' for col in diff_cols]
    df.drop(differenced_names, axis=1, inplace=True)

    return result



def create_differenced_variables(df, diff_cols, id_col='TICKER', date_col='Date', difference_lag=1):
    """
    Note: partially inplace
    """
    df.sort_values([id_col, date_col], inplace=True)
    df = add_missing_group_rows(df, [id_col, date_col])

    for col in diff_cols:
        _create_differenced_variable(df, col, id_col=id_col, difference_lag=difference_lag)

    df = drop_missing_group_rows(df, [id_col, date_col])

    return df


def _create_differenced_variable(df, col, id_col='TICKER', difference_lag=1, keep_lag=False):
    """
    Note: inplace
    """
    df[col + '_lag'] = df.groupby(id_col)[col].shift(difference_lag)
    df[col + ' Change'] = df[col] - df[col + '_lag']

    if not keep_lag:
        df.drop(col + '_lag', axis=1, inplace=True)

def _convert_variable_names(yvar, xvars, diff_cols):
    if yvar in diff_cols:
        yvar = yvar + ' Change'

    out_xvars = _convert_list_of_variables_to_difference_names(xvars, diff_cols)

    return yvar, out_xvars

def _convert_list_of_variables_to_difference_names(varlist, diff_cols):

    # if 'all' or 'xvars' is passed, no conversion needed
    if _is_special_lag_keyword(varlist):
        return varlist

    out_vars = []
    for var in varlist:
        if var in diff_cols:
            out_vars.append(var + ' Change')
        else:
            out_vars.append(var)
    return out_vars

def _convert_interaction_tuples(interaction_tuples, diff_cols):
    out_tuples = []
    for tup in interaction_tuples:
        out_tuples.append(tuple([var + ' Change' if var in diff_cols else var for var in tup]))

    return out_tuples


def _is_diff_reg_str(reg_str):
    return reg_str in ('diff', 'difference', 'diff_reg', 'diff reg', 'difference reg', 'difference regression')


from dero.latex.texgen import _usepackage_str


class Package:

    def __init__(self, name, modifier_str=None):
        self.name = name
        self.modifier_str = modifier_str

    def __repr__(self):
        return f'<Package(name={self.name})>'

    def __str__(self):
        return _usepackage_str(self.name, self.modifier_str)




import os
import shutil

from dero.latex.tools import date_time_move_latex


def _document_to_pdf_and_move(document, outfolder, image_paths=None, outname='figure', as_document=True,
                              move_folder_name='Figures'):

    # We will change paths, so save original to switch back to
    orig_path = os.getcwd()

    os.chdir(outfolder)

    # Create tex file
    outname_tex = outname + '.tex'
    with open(outname_tex, 'w') as f:
        f.write(str(document))

    if image_paths:
        # Copy first time for creation of pdf
        sources_tempfolder = os.path.join(outfolder, 'Sources')
        if not os.path.exists(sources_tempfolder):
            os.makedirs(sources_tempfolder)
        [_copy_if_needed(filepath, os.path.join(sources_tempfolder, _latex_valid_basename(filepath)))
         for filepath in image_paths]

    if as_document:
        os.system('pdflatex ' + '"' + outname_tex + '"') #create PDF
    new_outfolder = date_time_move_latex(outname, outfolder, folder_name=move_folder_name) #move table into appropriate date/number folder

    if image_paths and new_outfolder:
        # Copy second time to move pictures along with pdf
        sources_outfolder = os.path.join(new_outfolder, 'Sources')
        _move_if_needed(sources_tempfolder, sources_outfolder)

    os.chdir(orig_path)


def _copy_if_needed(src, dst):
    try:
        shutil.copy(src, dst)
    except shutil.SameFileError:
        pass


def _move_if_needed(src, dst):
    try:
        shutil.move(src, dst)
    except shutil.SameFileError:
        pass


def _move_folder_or_move_files_if_destination_folder_exists(src, dst):
    try:
        _move_if_needed(src, dst)
    except shutil.Error:
        files = [file for file in next(os.walk(src))[2]]
        [_move_if_needed(file, dst) for file in files]


def _latex_valid_basename(filepath):
    basename = os.path.basename(filepath)
    return basename.replace(' ', '_').replace('/','_').replace('%','pct')


from dero.latex.table.models.panels import Panel
from dero.latex.table.models.data.table import DataTable
from dero.latex.table.models.labels.table import LabelCollection, LabelTable
from dero.latex.table.models.labels.label import Label
from dero.latex.table.models.table.table import Table
from dero.latex.logic.tools import show_contents


from dero.latex.models.environment import Environment
from dero.latex.models.item import Item
from dero.latex.texgen import _document_class_str

from dero.latex.models.landscape import Landscape

class DocumentEnvironment(Environment):
    name = 'document'

    def __init__(self):
        super().__init__(name=self.name)

class Document(Item):
    name = 'document'

    def __init__(self, packages, content, landscape=False):
        from dero.latex.logic.builder import _build
        self.packages = packages

        pre_env_contents = _build([
            _document_class_str(),
            *[str(package) for package in self.packages]
        ])

        if landscape:
            content = Landscape().wrap(str(content))

        super().__init__(self.name, content, pre_env_contents=pre_env_contents)

    def __repr__(self):
        return f'<Document>'


from dero.latex.models.package import Package
from dero.latex.table.models.texgen.newcolumntype import NewColumnTypes

_default_package_strs = [
    'amsmath',
    'pdflscape',
    'dcolumn',
    'booktabs',
    'array',
    'threeparttable'
]

_direct_default_packages = [
    Package('geometry', modifier_str='margin=0.3in')
]

default_packages = [Package(package_str) for package_str in _default_package_strs] + \
                   _direct_default_packages + [NewColumnTypes()]


from dero.reg.differenced import _is_diff_reg_str
from dero.reg.lag.create import varname_to_lagged_varname


def _set_regressor_order(regressor_order, reg_kwargs):

    # processing needed for lagged regression
    if ('num_lags' in reg_kwargs) or ('lag_tuple' in reg_kwargs):
        regressor_order = convert_regressor_order_for_lags(regressor_order, reg_kwargs)

    # processing needed for difference regression
    if ('reg_type' in reg_kwargs) and (_is_diff_reg_str(reg_kwargs['reg_type'])):
        regressor_order = convert_regressor_order_for_diff(regressor_order, reg_kwargs)

    return regressor_order

def convert_regressor_order_for_diff(regressor_order, reg_kwargs):
    if 'diff_cols' in reg_kwargs:
        cols = reg_kwargs['diff_cols']
    else:
        cols = 'all'

    return _convert_regressor_order_for_diff(regressor_order, cols)

def _convert_regressor_order_for_diff(regressor_order, cols='all'):
    if cols == 'all':
        cols = regressor_order.copy()

    return [col + ' Change' if col in cols else col for col in regressor_order]

def convert_regressor_order_for_lags(regressor_order, reg_kwargs):

    # Set lag variables
    if 'lag_variables' in reg_kwargs:
        lag_cols = reg_kwargs['lag_variables']
    else:
        lag_cols = 'all'

    if lag_cols in ('xvars', 'all'):
        lag_cols = regressor_order.copy()

    # Set number of lags
    if 'num_lags' in reg_kwargs:
        # Here a single number of lags has been passed
        lags = [reg_kwargs['num_lags']]
    elif 'lag_tuple' in reg_kwargs:
        # Here multiple lags are being passed in a tuple
        lags = reg_kwargs['lag_tuple']

    return _convert_regressor_order_for_lags(regressor_order, lag_cols, lags)

def _convert_regressor_order_for_lags(regressor_order, lag_cols, lags):
    out_cols = []
    for col in regressor_order:
        if col in lag_cols:
            [out_cols.append(varname_to_lagged_varname(col, num_lags=num_lags)) for num_lags in lags]
        else:
            out_cols.append(col)

    return out_cols